import json
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Union

from marsbots.util import pythonify_json


class SettingsManager(ABC):
    def __init__(self, settings: dict, defaults) -> None:
        self.settings = settings
        self.defaults = defaults

    @property
    @abstractmethod
    def is_initialized(self):
        raise NotImplementedError()

    @abstractmethod
    def initialize(self):
        raise NotImplementedError()

    @abstractmethod
    def load_settings(self):
        raise NotImplementedError()

    @abstractmethod
    def update_settings(self, updates):
        raise NotImplementedError()

    def initialize_guild(self, guild_id):
        self.settings[guild_id] = {}
        guild_settings = self.settings[guild_id]
        print("###########")
        print(self.defaults)
        for k, v in self.defaults.__dict__.items():
            guild_settings[k] = {}
            guild_settings[k]["default"] = v
            guild_settings[k]["channels"] = {}
        self.update_settings()
        return guild_settings

    def get_guild_settings(self, guild_id):
        guild_settings = self.settings.get(guild_id)
        if not guild_settings:
            guild_settings = self.initialize_guild(guild_id)
        return guild_settings

    def get_setting(self, guild_id, setting_name):
        guild_settings = self.get_guild_settings(guild_id)
        return guild_settings[setting_name]["default"]

    def get_channel_setting(self, channel_id, guild_id, setting_name):
        guild_settings = self.get_guild_settings(guild_id)
        setting = guild_settings[setting_name]
        channel_setting = setting["channels"].get(channel_id)
        return channel_setting if channel_setting else setting["default"]

    def update_setting(self, guild_id, setting_name, value):
        guild_settings = self.get_guild_settings(guild_id)
        guild_settings[setting_name]["default"] = value
        self.update_settings()

    def update_channel_setting(self, channel_id, guild_id, setting_name, value):
        guild_settings = self.get_guild_settings(guild_id)
        guild_settings[setting_name]["channels"][channel_id] = value
        self.update_settings()


class LocalSettingsManager(SettingsManager):
    def __init__(self, filepath: Union[Path, str], defaults) -> None:
        self.filepath = Path(filepath)
        self.settings = self.load_settings()
        super().__init__(self.settings, defaults)

    @property
    def is_initialized(self):
        if self.filepath.exists():
            return True
        return False

    def initialize(self):
        self.filepath.write_text(json.dumps({}))

    def load_settings(self):
        if not self.is_initialized:
            self.initialize()
        return pythonify_json(json.loads(self.filepath.read_text()))

    def update_settings(self):
        with open(self.filepath, "w") as f:
            json.dump(self.settings, f)


class MongoSettingsManager(SettingsManager):
    pass



import asyncio
import io
import json

import aiohttp
import discord
import requests

from marsbots.resources.discord_utils import update_message


async def generation_loop(
    gateway_url,
    minio_url,
    interaction_message,
    config,
    refresh_interval: int,
):
    result = requests.post(gateway_url + "/request_creation", json=config)

    if not await check_server_result_ok(result, interaction_message):
        return

    result = json.loads(result.content)
    task_id = result["task_id"]
    current_sha = None
    await update_progress(interaction_message, 0)

    while True:
        result = requests.post(
            gateway_url + "/get_creations",
            json={"task_id": task_id},
        )

        if not await check_server_result_ok(result, interaction_message):
            return

        result = json.loads(result.content)

        if not result:
            message_suffix = "_Server error: task ID not found_"
            message_content = appender(interaction_message.content, message_suffix)
            return await update_message(interaction_message, content=message_content)

        result = result[0]
        status = result["status"]
        progress = -1 if status == "pending" else result["status_code"]
        await update_progress(interaction_message, progress)

        if "intermediate_sha" in result:
            last_sha = result["intermediate_sha"][-1]
            if last_sha != current_sha:
                current_sha = last_sha
                sha_url = f"{minio_url}/{current_sha}"
                filename = f"{current_sha}.png"
                discord_file = await get_discord_file_from_url(sha_url, filename)
                await update_image(interaction_message, discord_file)

        if status == "failed":
            message_suffix = "_Server error: Eden task failed_"
            message_content = appender(interaction_message.content, message_suffix)
            return await update_message(interaction_message, content=message_content)

        if status not in ["pending", "running"]:
            break

        await asyncio.sleep(refresh_interval)


def appender(message, suffix):
    return message.split("\n")[0] + "\n\n" + suffix


async def update_progress(message, progress):
    if progress == "__none__":
        progress = 0
    progress_str = "pending" if progress == -1 else f"**{progress}%** complete"
    message_suffix = "" if progress == 100 else f"_Creation is {progress_str}_"
    message_content = appender(message.content, message_suffix)
    await update_message(message, content=message_content)


async def update_queue_position(message, position):
    message_suffix = f"_Queue position: **{position}**_"
    message_content = appender(message.content, message_suffix)
    await update_message(message, content=message_content)


async def update_image(message, image):
    await update_message(message, files=[image])


async def get_discord_file_from_url(url, filename):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            if resp.status != 200:
                return None
            data = io.BytesIO(await resp.read())
            discord_file = discord.File(data, filename)
            return discord_file


async def check_server_result_ok(result, interaction_message):
    if result.status_code != 200:
        error_message = result.content.decode("utf-8")
        message_suffix = f"_Server error: {error_message}_"
        message_content = appender(interaction_message.content, message_suffix)
        await update_message(interaction_message, content=message_content)
    return result.status_code == 200



import logging
import re
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import List
from typing import Optional

import discord
from discord.ext import commands

from marsbots import constants
from marsbots.exceptions import ChannelNotFoundError


def is_mentioned(message: discord.Message, user: discord.User) -> bool:
    """
    Checks if a user is mentioned in a message.
    :param message: The message to check.
    :param user: The user to check.
    :return: True if the user is mentioned, False otherwise.
    """
    return user.id in [m.id for m in message.mentions]


def role_is_mentioned(message: discord.Message, role_name: str) -> bool:
    return role_name in [r.name for r in message.role_mentions]


async def process_mention_as_command(
    ctx: str,
    cog: commands.Cog,
    command_not_found_response: str = constants.COMMAND_NOT_FOUND_MESSAGE,
):
    try:
        split_message = ctx.message.content.split(" ")
        message_text, args = split_message[1], split_message[2:]
        cmd = getattr(cog, message_text)
        await cmd(ctx, *args)
    except Exception as e:
        logging.error(e)
        await ctx.message.channel.send(command_not_found_response)


async def get_discord_messages(
    channel: discord.TextChannel,
    limit: int,
    after: Optional[timedelta] = None,
) -> list:
    """
    Gets the last x messages from a channel.
    :param channel: The channel to get the messages from.
    :param limit: The number of messages to get.
    :return: The last x messages from the channel.
    """
    if after is not None:
        time = datetime.now(timezone.utc) - after
    else:
        time = None

    raw_messages = await channel.history(
        limit=limit,
        oldest_first=False,
        after=time,
    ).flatten()

    raw_messages.reverse()

    return raw_messages


def replace_bot_mention(
    message_text: str,
    only_first: bool = True,
    replacement_str: str = "",
) -> str:
    """
    Removes all mentions from a message.
    :param message: The message to remove mentions from.
    :return: The message with all mentions removed.
    """
    if only_first:
        return re.sub(r"<@\d+>", replacement_str, message_text, 1)
    else:
        return re.sub(r"<@\d+>", replacement_str, message_text)


def remove_role_mentions(message_text: str) -> str:
    """
    Removes all role mentions from a message.
    :param message_text: The message to remove role mentions from.
    :return: The message with all role mentions removed.
    """
    return re.sub(r"<@&\d+>", "", message_text)


async def get_reply_chain(
    ctx,
    message: discord.Message,
    depth: int,
) -> List[discord.Message]:
    messages = []
    count = 0
    while message and message.reference and count < depth:
        message = await ctx.fetch_message(message.reference.message_id)
        messages.append(message)
        count += 1
    messages.reverse()
    return messages


def replace_mentions_with_usernames(message_content: str, mentions) -> str:
    """
    Replaces all mentions with their usernames.
    :param message_content: The message to replace mentions in.
    :return: The message with all mentions replaced with their usernames.
    """
    for mention in mentions:
        message_content = message_content.replace(f"<@{mention.id}>", mention.name)
    return message_content


def in_channels(channel_ids: List[int]) -> commands.check:
    async def predicate(ctx):
        if ctx.channel.id in channel_ids:
            return True
        print("command not valid in channel")

    return commands.check(predicate)


async def wait_for_user_reply(bot: commands.Bot, sender_id: int) -> discord.Message:
    """
    Waits for a user to reply to a message.
    :return: The message the user replied to.
    """

    def check(message: discord.Message) -> bool:
        return sender_id == message.author.id

    return await bot.wait_for("message", check=check)


async def update_message(
    message: discord.Message,
    content: Optional[str] = None,
    files: Optional[List[discord.File]] = None,
    clear_previous_images: bool = True,
) -> discord.Message:
    if content:
        await message.edit(content=content)
    if files:
        if clear_previous_images:
            await message.edit(files=files, attachments=[])
        else:
            await message.edit(files=files)


def filter_application_command_messages(
    messages: List[discord.Message],
) -> List[discord.Message]:
    return [m for m in messages if m.type != discord.MessageType.application_command]


def get_channel_id_by_channel_name(channel_name: str, ctx) -> int:
    try:
        channel = next(
            filter(lambda x: x.name == channel_name, ctx.guild.text_channels),
        )
        return channel.id
    except StopIteration:
        raise ChannelNotFoundError("No text channel found matching the given name.")


def get_nick(member):
    if hasattr(member, "nick") and member.nick is not None:
        return member.nick
    else:
        return member.name



from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent
DOTENV_PATH = ROOT_DIR / ".env"
LOG_DIR = ROOT_DIR / "logs"

COMMAND_NOT_FOUND_MESSAGE = "Command not found."

UNLIKELY_PREFIX = "fhkjdjksfjkdhkjfhksdjfsdjkfbwereuifbiudbyuifvyudsvyu"

CHANNEL_NOT_FOUND_MESSAGE = "Could not find a text channel matching that name."



from discord.ext import commands

from marsbots import config
from marsbots.resources.language_models import OpenAIGPT3LanguageModel


class ExampleCog2(commands.Cog):
    def __init__(self, bot: commands.bot) -> None:
        self.bot = bot
        self.language_model = OpenAIGPT3LanguageModel(config.LM_OPENAI_API_KEY)

    @commands.command()
    async def example(self, ctx: commands.context) -> None:
        await ctx.send("Hello world from cog 2!")


def setup(bot: commands.Bot) -> None:
    bot.add_cog(ExampleCog2(bot))



import asyncio
import os
from dataclasses import dataclass
from pathlib import Path

import discord
from discord.commands import slash_command
from discord.ext import commands

from marsbots import config
from marsbots.models import ChatMessage
from marsbots.programs.lm import complete_text
from marsbots.resources.discord_utils import get_discord_messages
from marsbots.resources.discord_utils import in_channels
from marsbots.resources.discord_utils import update_message
from marsbots.resources.discord_utils import wait_for_user_reply
from marsbots.resources.language_models import OpenAIGPT3LanguageModel
from marsbots.resources.settings_manager import LocalSettingsManager


class ButtonView(discord.ui.View):
    def __init__(self):
        # making None is important if you want the button work after restart!
        super().__init__(timeout=None)

    # custom_id is required and should be unique for <commands.Bot.add_view>
    # attribute emoji can be used to include emojis which can be default str emoji or str(<:emojiName:int(ID)>)
    # timeout can be used if there is a timeout on the button interaction. Default timeout is set to 180.
    @discord.ui.button(
        style=discord.ButtonStyle.blurple,
        custom_id="counter:firstButton",
        label="Button",
    )
    async def leftButton(self, button, interaction):
        await interaction.response.edit_message(content="button was pressed!")


@dataclass
class ExampleBotSettings:
    setting1: int = 10
    setting2: int = 10


class ExampleCog(commands.Cog):
    def __init__(self, bot: commands.bot) -> None:
        self.bot = bot
        self.language_model = OpenAIGPT3LanguageModel(config.LM_OPENAI_API_KEY)
        self.settings_path = Path("./examplebot_settings.json")
        self.settings_manager = LocalSettingsManager(
            self.settings_path,
            defaults=ExampleBotSettings(),
        )

    @commands.command()
    async def get_commands(self, ctx) -> None:
        print([c.qualified_name for c in self.walk_commands()])

    @commands.command()
    async def whereami(self, ctx) -> None:
        await ctx.send("Hello from a custom cog")
        await ctx.send(ctx.guild.id)

    @slash_command(guild_ids=[config.TEST_GUILD_ID])
    async def howami(self, ctx) -> None:
        await ctx.respond("doing great")

    @commands.command()
    async def get_messages(self, ctx: commands.Context) -> None:
        messages = await get_discord_messages(ctx.channel, 10)
        for message in messages:
            msg = ChatMessage(
                content=message.content,
                sender=message.author.name,
            )
            print(msg)

    @commands.command()
    async def complete(
        self,
        ctx: commands.context,
        max_tokens: int,
        *input_text: str,
    ) -> None:
        prompt = " ".join(input_text)
        async with ctx.channel.typing():
            completion = complete_text(self.language_model, prompt, max_tokens)
            await ctx.send(prompt + completion)

    @slash_command(guild_ids=[config.TEST_GUILD_ID])
    async def complete_some_text(
        self,
        ctx,
        max_tokens: int,
        prompt: str,
    ) -> None:
        completion = await complete_text(self.language_model, prompt, max_tokens)
        print(prompt + completion)
        await ctx.respond(prompt + completion)

    @slash_command(
        guild_ids=[config.TEST_GUILD_ID],
        name="slash_command_name",
        description="command description!",
    )
    async def button(self, ctx):
        navigator = ButtonView()
        await ctx.respond("press the button.", view=navigator)

    @commands.command()
    async def resolve(self, ctx, message_id):
        msg = await ctx.fetch_message(message_id)
        print(msg.content)

    @commands.command()
    @in_channels([config.TEST_CHANNEL_ID])
    async def test_in_channels(self, ctx):
        await ctx.send("In the test channel.")

    @commands.command()
    async def test_edit_message(self, ctx):
        assets_path = os.path.join(os.path.dirname(__file__), "assets")
        filepaths = [
            os.path.join(assets_path, fname)
            for fname in ["be-patient.png", "s-l1600.jpg", "uc2.png", "s-l1600 (1).jpg"]
        ]
        files = [discord.File(filepath) for filepath in filepaths[:2]]
        message = await ctx.send("Hey", files=files)
        await asyncio.sleep(3)
        await update_message(message, content="Goodbye", image_paths=filepaths[2:])

    @commands.command()
    async def get_setting_1(self, ctx):
        setting = self.settings_manager.get_setting(ctx.guild.id, "setting1")
        await ctx.send(setting)

    @commands.command()
    async def get_channel_setting_1(self, ctx):
        setting = self.settings_manager.get_channel_setting(
            ctx.channel.id,
            ctx.guild.id,
            "setting1",
        )
        await ctx.send(setting)

    @commands.command()
    async def get_settings(self, ctx):
        print(self.settings_manager.settings)

    @commands.command()
    async def update_setting_1(self, ctx, value):
        self.settings_manager.update_setting(ctx.guild.id, "setting1", value)
        await ctx.send("updated setting1")

    @commands.command()
    async def update_channel_setting_1(self, ctx, value):
        self.settings_manager.update_channel_setting(
            ctx.channel.id,
            ctx.guild.id,
            "setting1",
            value,
        )
        await ctx.send("updated channel setting1")

    @slash_command(guild_ids=[config.TEST_GUILD_ID])
    async def update_settings(
        self,
        ctx,
        setting: discord.Option(
            str,
            description="Setting name to update",
            required=True,
            choices=list(ExampleBotSettings.__dataclass_fields__.keys()),
        ),
        channel_name: discord.Option(
            str,
            description="Channel to update setting for",
            required=False,
        ),
    ):

        if channel_name:
            await self.handle_update_channel_settings(ctx, setting, channel_name)
        else:
            await self.handle_update_settings(ctx, setting)

    async def handle_update_settings(self, ctx, setting):
        await ctx.respond(
            f"Enter a new value for {setting}. (Currently"
            f" {self.settings_manager.get_setting(ctx.guild.id, setting)})",
        )
        resp = await wait_for_user_reply(self.bot, ctx.author.id)
        try:
            new_val = ExampleBotSettings.__dataclass_fields__[setting].type(
                resp.content,
            )
        except ValueError:
            await ctx.send(f"{resp.content} is not a valid value for {setting}")
            return
        self.settings_manager.update_setting(ctx.guild.id, setting, new_val)
        await ctx.send(f"Updated {setting} to {new_val}")

    async def handle_update_channel_settings(self, ctx, setting, channel_name):
        channel = discord.utils.get(ctx.guild.channels, name=channel_name)
        if not channel:
            await ctx.respond(f"No channel named {channel_name}")
            return

        await ctx.respond(
            f"Enter a new value for {setting}. (Currently"
            f" {self.settings_manager.get_channel_setting(channel.id, ctx.guild.id, setting)})",
        )
        resp = await wait_for_user_reply(self.bot, ctx.author.id)
        try:
            new_val = ExampleBotSettings.__dataclass_fields__[setting].type(
                resp.content,
            )
        except ValueError:
            await ctx.send(f"{resp.content} is not a valid value for {setting}")
            return
        self.settings_manager.update_channel_setting(
            channel.id,
            ctx.guild.id,
            setting,
            new_val,
        )
        await ctx.send(f"Updated {setting} to {new_val}")


def setup(bot: commands.Bot) -> None:
    bot.add_cog(ExampleCog(bot))



import discord
from discord.ext import commands

from . import prompts
from marsbots import config
from marsbots.models import ChatMessage
from marsbots.programs.lm import complete_text
from marsbots.resources.discord_utils import get_discord_messages
from marsbots.resources.discord_utils import get_reply_chain
from marsbots.resources.discord_utils import is_mentioned
from marsbots.resources.discord_utils import replace_bot_mention
from marsbots.resources.discord_utils import replace_mentions_with_usernames
from marsbots.resources.language_models import OpenAIGPT3LanguageModel


class ExampleCog(commands.Cog):
    def __init__(self, bot: commands.bot) -> None:
        self.bot = bot
        self.language_model = OpenAIGPT3LanguageModel(config.LM_OPENAI_API_KEY)

        self.name = "Chatbot"
        self.should_follow_reply_chain = True

    @commands.Cog.listener("on_message")
    async def on_message(self, message: discord.Message) -> None:
        if is_mentioned(message, self.bot.user):
            ctx = await self.bot.get_context(message)
            async with ctx.channel.typing():
                prompt = await self.format_prompt(ctx, message)
                completion = await complete_text(
                    self.language_model,
                    prompt,
                    max_tokens=80,
                    stop=["**[", "\n\n"],
                )
                await message.reply(completion)

    async def format_prompt(
        self,
        ctx: commands.context,
        message: discord.Message,
    ) -> str:
        last_messages = await get_discord_messages(ctx.channel, 1)
        reply_chain = await get_reply_chain(ctx, message, depth=6)
        if reply_chain:
            reply_chain = self.format_reply_chain(reply_chain)
        last_message_text = str(
            ChatMessage(
                f"{self.message_preprocessor(last_messages[0])}",
                message.author.display_name,
            ),
        ).strip()
        prompt = prompts.PREFIX

        if reply_chain and self.should_follow_reply_chain:
            prompt += f"{reply_chain}\n"

        prompt += "\n".join(
            [
                last_message_text,
                f"**[{self.name}]**:",
            ],
        )

        print(prompt)

        return prompt

    def format_reply_chain(self, messages):
        reply_chain = []
        for message in messages:
            if message.author.id == self.bot.user.id:
                sender_name = self.name
            else:
                sender_name = message.author.display_name
            reply_chain.append(
                ChatMessage(
                    content=f"{self.message_preprocessor(message)}",
                    sender=sender_name,
                ),
            )
        return "\n".join([str(message).strip() for message in reply_chain])

    def message_preprocessor(self, message: discord.Message) -> str:
        message_content = replace_bot_mention(message.content, only_first=True)
        message_content = replace_mentions_with_usernames(
            message_content,
            message.mentions,
        )
        message_content = message_content.strip()
        return message_content


def setup(bot: commands.Bot) -> None:
    bot.add_cog(ExampleCog(bot))
